
/*
 * Copyright (C) 1998,  The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Visualization
 *      Group at Lawrence Berkeley National Laboratory.
 * 4. Neither the name of the University of California, Berkeley nor of the 
 *    Lawrence Berkeley National Laboratory may be used to endorse or 
 *    promote products derived from this software without specific prior 
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * This work is supported by the U. S. Department of Energy under contract 
 * number DE-AC03-76SF00098 between the U. S. Department of Energy and the 
 * University of California.
 *
 *	Author: Wes Bethel
 *		Lawrence Berkeley National Laboratory
 *              Berkeley, California 
 *
 *  "this software is 100% hand-crafted by a human being in the USA"
 */
 

#include "inet.h"
#include "hsocket.h"

#include <avs/avs.h>
#include <avs/field.h>


void
to_data_size(char *msg,
	     int dims[5])
{
    void *config_xdr_buf;
    XDR config_xdr;
    int xdr_buf_size;		/* bytes */

    /*
     * take a raw stream of bytes which is an XDR represenation
     * of an array of 5 int's, and turn them into an array of ints
     * for the local machine.
     *
     * since some architectures don't support the
     * xdr_sizeof() call, it's value is hard-coded in a header file. we
     * should be safe for a while.
     */


    config_xdr_buf = msg;

    xdrmem_create(&config_xdr,
		  (const caddr_t)(config_xdr_buf),
		  SIZEOF_XDR_INT*5, /* a priori knowledge */
		  XDR_DECODE);
    
    xdr_vector(&config_xdr,
	       (caddr_t)dims,
	       (const u_int)5,
	       (const u_int)sizeof(int),
	       xdr_int);

    free(config_xdr_buf);
    xdr_destroy(&config_xdr);

}

void
get_data(char *msg,
	 double *data,
	 int ndata)
{
    /*
     * take a raw stream of bytes which is an array of XDR doubles, and
     * turn them into an array of doubles on the local architecture. the
     * length of the raw byte stream is a function of the size of the
     * overall data array (sent in a previous msg) and the size of
     * an XDR double.
     *
     * since some architectures don't support the
     * xdr_sizeof() call, it's value is hard-coded in a header file. we
     * should be safe for a while.
     */
    void *data_xdr_buf;
    XDR data_xdr;
    int data_buf_size;		/* bytes */


    data_xdr_buf = msg;

    xdrmem_create(&data_xdr,
		  (const caddr_t)(data_xdr_buf),
		  SIZEOF_XDR_DOUBLE*ndata, /* a priori knowledge */
		  XDR_DECODE);
    
    xdr_vector(&data_xdr,
	       (caddr_t)data,
	       (const u_int)ndata,
	       (const u_int)sizeof(double),
	       xdr_double);

    free(data_xdr_buf);
    xdr_destroy(&data_xdr);
}


int
DataRecvDesc()
{
    /*
     * this routine builds the UI on the AVS module. the only
     * thing that can be controlled is the Socket Port number.
     */
    int p;

    AVSset_module_name("Data Receiver",MODULE_DATA);
    AVScreate_output_port("output field", "field");
    
    p = AVSadd_parameter("dummy1","string","Host/Network Configuration","","");
    AVSconnect_widget(p,"text");
    AVSadd_parameter_prop(p,"width","integer",4);

    p = AVSadd_parameter("port num","integer",SERV_TCP_PORT,4097,65535);
    AVSconnect_widget(p,"typein integer");

    
}


void
make_new_field(AVSfield **outfield,
	       int dims[5])
{
    /*
     * build a new AVS field containing scalar doubles of the
     * requested size.
     */
    
    AVSfield template;
    int i,ndim;

    for (ndim=0,i=0;i<5;i++)
	if (dims[i] > 0)
	    ndim++;

    memset(&template,0,sizeof(AVSfield));
    template.ndim = ndim;
    template.nspace = ndim;
    template.veclen = 1;
    template.size = sizeof(double);
    template.uniform = UNIFORM;
    template.type = AVS_TYPE_DOUBLE;

    if (*outfield)
	AVSfield_free(*outfield);

    *outfield = (AVSfield *)AVSfield_alloc(&template,dims);
}

main(int ac,
     char *av[])
{

    
    SocketFD server_info, client_info;
    
    int status;
    char *msg;
    int msg_length;
    int command_opcode;
    char command[8];
    int dims[5];
    int ndata,i;
    double *data;
    timeout_thingy timeout = {1000,0};

    /* AVS variables */
    AVSfield *outfield=NULL;
    char *avs_dummy1;
    int avs_port_num;

    AVScorout_init(ac,av,DataRecvDesc);

    AVScorout_input(&avs_dummy1,&avs_port_num);

    /*
     * loop forever. the port number is only looked at once when
     * this module is first started. to change the port number, you'll
     * have to kill the module and start over.
     */

    status = open_server_socket(&server_info,avs_port_num);
    for (;;)
    {

	fprintf(stderr," waiting for connection...");
	status = server_waitfor_connection(&server_info,&client_info,NULL);
    
	for (;status==1;)
	{
	    fprintf(stderr, "\n\n ***********  new data ************ \n\n");
	
	    server_receive_message(&client_info,command,&msg,&msg_length,&timeout);
	    if (strcmp(command,"clos") == 0)
		break;
	
	    to_data_size(msg,dims);	

	    fprintf(stderr,"size of array is %d %d %d %d %d \n",dims[0],dims[1],dims[2],dims[3],dims[4]);

	    for (i=0,ndata=1;i<5;i++)
		ndata *= (dims[i] == 0.) ? 1 : dims[i];

	    make_new_field(&outfield,dims);
	    
	    data = (double *)malloc(sizeof(double)*ndata);

	    server_receive_message(&client_info,command,&msg,&msg_length,&timeout);
	    if (strcmp(command,"clos") == 0)
		break;

	    get_data(msg,data,ndata);
	    
	    memcpy(outfield->field_union.field_data_double_u,
		   data,sizeof(double)*ndata);

	    AVScorout_output(outfield);
	    AVScorout_exec(); 
	    
	    status = 1;
	    free((void *)data);
	    
	}
	fprintf(stderr," now closing it.\n");
	server_close_connection(&client_info);
    }

}
